@@ -3,7 +3,7 @@
   Copyright (C) 2008       SUSE Linux Products GmbH
   Copyright (C) 2008       Tejun Heo <teheo@suse.de>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
@@ -37,8 +37,6 @@
  */
 
 
-#include <config.h>
-
 #include <sys/types.h>
 #include <fcntl.h>
 #include <sys/stat.h>
@@ -47,6 +45,7 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include <errno.h>
+#include <unistd.h>
 #include "ioctl.h"
 
 const char *usage =
@@ -123,16 +122,17 @@ int main(int argc, char **argv)
 		if (!argc) {
 			if (ioctl(fd, FIOC_GET_SIZE, &size)) {
 				perror("ioctl");
-				return 1;
+				goto error;
 			}
 			printf("%zu\n", size);
 		} else {
 			size = param[0];
 			if (ioctl(fd, FIOC_SET_SIZE, &size)) {
 				perror("ioctl");
-				return 1;
+				goto error;
 			}
 		}
+		close(fd);
 		return 0;
 
 	case 'r':
@@ -140,13 +140,18 @@ int main(int argc, char **argv)
 		rc = do_rw(fd, cmd == 'r', param[0], param[1],
 			   &prev_size, &new_size);
 		if (rc < 0)
-			return 1;
+			goto error;
 		fprintf(stderr, "transferred %d bytes (%zu -> %zu)\n",
 			rc, prev_size, new_size);
+		close(fd);
 		return 0;
 	}
 
- usage:
+usage:
 	fprintf(stderr, "%s", usage);
 	return 1;
+
+error:
+	close(fd);
+	return 1;
 }
