@@ -2,7 +2,7 @@
   FUSE: Filesystem in Userspace
   Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
@@ -18,9 +18,7 @@
  * \include hello_ll.c
  */
 
-#define FUSE_USE_VERSION 31
-
-#include <config.h>
+#define FUSE_USE_VERSION FUSE_MAKE_VERSION(3, 12)
 
 #include <fuse_lowlevel.h>
 #include <stdio.h>
@@ -55,6 +53,18 @@ static int hello_stat(fuse_ino_t ino, struct stat *stbuf)
 	return 0;
 }
 
+static void hello_ll_init(void *userdata, struct fuse_conn_info *conn)
+{
+	(void)userdata;
+
+	/* Disable the receiving and processing of FUSE_INTERRUPT requests */
+	conn->no_interrupt = 1;
+
+	/* Test setting flags the old way */
+	conn->want = FUSE_CAP_ASYNC_READ;
+	conn->want &= ~FUSE_CAP_ASYNC_READ;
+}
+
 static void hello_ll_getattr(fuse_req_t req, fuse_ino_t ino,
 			     struct fuse_file_info *fi)
 {
@@ -140,7 +150,7 @@ static void hello_ll_open(fuse_req_t req, fuse_ino_t ino,
 {
 	if (ino != 2)
 		fuse_reply_err(req, EISDIR);
-	else if ((fi->flags & 3) != O_RDONLY)
+	else if ((fi->flags & O_ACCMODE) != O_RDONLY)
 		fuse_reply_err(req, EACCES);
 	else
 		fuse_reply_open(req, fi);
@@ -155,12 +165,64 @@ static void hello_ll_read(fuse_req_t req, fuse_ino_t ino, size_t size,
 	reply_buf_limited(req, hello_str, strlen(hello_str), off, size);
 }
 
-static struct fuse_lowlevel_ops hello_ll_oper = {
-	.lookup		= hello_ll_lookup,
-	.getattr	= hello_ll_getattr,
-	.readdir	= hello_ll_readdir,
-	.open		= hello_ll_open,
-	.read		= hello_ll_read,
+static void hello_ll_getxattr(fuse_req_t req, fuse_ino_t ino, const char *name,
+							  size_t size)
+{
+	(void)size;
+	assert(ino == 1 || ino == 2);
+	if (strcmp(name, "hello_ll_getxattr_name") == 0)
+	{
+		const char *buf = "hello_ll_getxattr_value";
+		fuse_reply_buf(req, buf, strlen(buf));
+	}
+	else
+	{
+		fuse_reply_err(req, ENOTSUP);
+	}
+}
+
+static void hello_ll_setxattr(fuse_req_t req, fuse_ino_t ino, const char *name,
+							  const char *value, size_t size, int flags)
+{
+	(void)flags;
+	(void)size;
+	assert(ino == 1 || ino == 2);
+	const char* exp_val = "hello_ll_setxattr_value";
+	if (strcmp(name, "hello_ll_setxattr_name") == 0 &&
+	    strlen(exp_val) == size &&
+	    strncmp(value, exp_val, size) == 0)
+	{
+		fuse_reply_err(req, 0);
+	}
+	else
+	{
+		fuse_reply_err(req, ENOTSUP);
+	}
+}
+
+static void hello_ll_removexattr(fuse_req_t req, fuse_ino_t ino, const char *name)
+{
+	assert(ino == 1 || ino == 2);
+	if (strcmp(name, "hello_ll_removexattr_name") == 0)
+	{
+		fuse_reply_err(req, 0);
+	}
+	else
+	{
+		fuse_reply_err(req, ENOTSUP);
+	}
+}
+
+static const struct fuse_lowlevel_ops hello_ll_oper = {
+	.init = hello_ll_init,
+	.lookup = hello_ll_lookup,
+	.getattr = hello_ll_getattr,
+	.readdir = hello_ll_readdir,
+	.open = hello_ll_open,
+	.read = hello_ll_read,
+	.setxattr = hello_ll_setxattr,
+	.getxattr = hello_ll_getxattr,
+	.removexattr = hello_ll_removexattr,
 };
 
 int main(int argc, char *argv[])
@@ -168,6 +230,7 @@ int main(int argc, char *argv[])
 	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
 	struct fuse_session *se;
 	struct fuse_cmdline_opts opts;
+	struct fuse_loop_config *config;
 	int ret = -1;
 
 	if (fuse_parse_cmdline(&args, &opts) != 0)
@@ -185,6 +248,13 @@ int main(int argc, char *argv[])
 		goto err_out1;
 	}
 
+	if(opts.mountpoint == NULL) {
+		printf("usage: %s [options] <mountpoint>\n", argv[0]);
+		printf("       %s --help\n", argv[0]);
+		ret = 1;
+		goto err_out1;
+	}
+
 	se = fuse_session_new(&args, &hello_ll_oper,
 			      sizeof(hello_ll_oper), NULL);
 	if (se == NULL)
@@ -201,8 +271,14 @@ int main(int argc, char *argv[])
 	/* Block until ctrl+c or fusermount -u */
 	if (opts.singlethread)
 		ret = fuse_session_loop(se);
-	else
-		ret = fuse_session_loop_mt(se, opts.clone_fd);
+	else {
+		config = fuse_loop_cfg_create();
+		fuse_loop_cfg_set_clone_fd(config, opts.clone_fd);
+		fuse_loop_cfg_set_max_threads(config, opts.max_threads);
+		ret = fuse_session_loop_mt(se, config);
+		fuse_loop_cfg_destroy(config);
+		config = NULL;
+	}
 
 	fuse_session_unmount(se);
 err_out3:
