@@ -2,7 +2,7 @@
   FUSE: Filesystem in Userspace
   Copyright (C) 2016 Nikolaus Rath <Nikolaus@rath.org>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
@@ -58,9 +58,7 @@
  */
 
 
-#define FUSE_USE_VERSION 31
-
-#include <config.h>
+#define FUSE_USE_VERSION FUSE_MAKE_VERSION(3, 12)
 
 #include <fuse_lowlevel.h>
 #include <stdio.h>
@@ -72,26 +70,31 @@
 #include <stddef.h>
 #include <unistd.h>
 #include <pthread.h>
+#include <stdbool.h>
 
 /* We can't actually tell the kernel that there is no
    timeout, so we just send a big value */
 #define NO_TIMEOUT 500000
 
-/* We cannot check directly if e.g. O_RDONLY is set, since this is not
- * an individual bit (cf. open(2)) */
-#define ACCESS_MASK (O_RDONLY | O_WRONLY | O_RDWR)
-
 #define MAX_STR_LEN 128
 #define FILE_INO 2
 #define FILE_NAME "current_time"
 static char file_contents[MAX_STR_LEN];
 static int lookup_cnt = 0;
+static int open_cnt = 0;
 static size_t file_size;
+static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
 
 /* Keep track if we ever stored data (==1), and
    received it back correctly (==2) */
 static int retrieve_status = 0;
 
+static bool is_umount = false;
+
+/* updater thread tid */
+static pthread_t updater;
+
+
 /* Command line parsing */
 struct options {
     int no_notify;
@@ -129,6 +132,13 @@ static int tfs_stat(fuse_ino_t ino, struct stat *stbuf) {
     return 0;
 }
 
+static void tfs_init(void *userdata, struct fuse_conn_info *conn) {
+	(void)userdata;
+
+	/* Disable the receiving and processing of FUSE_INTERRUPT requests */
+	conn->no_interrupt = 1;
+}
+
 static void tfs_lookup(fuse_req_t req, fuse_ino_t parent,
                        const char *name) {
     struct fuse_entry_param e;
@@ -138,7 +148,6 @@ static void tfs_lookup(fuse_req_t req, fuse_ino_t parent,
         goto err_out;
     else if (strcmp(name, FILE_NAME) == 0) {
         e.ino = FILE_INO;
-        lookup_cnt++;
     } else
         goto err_out;
 
@@ -147,6 +156,18 @@ static void tfs_lookup(fuse_req_t req, fuse_ino_t parent,
     if (tfs_stat(e.ino, &e.attr) != 0)
         goto err_out;
     fuse_reply_entry(req, &e);
+
+    /*
+     * must only be set when the kernel knows about the entry,
+     * otherwise update_fs_loop() might see a positive count, but kernel
+     * would not have the entry yet
+     */
+    if (e.ino == FILE_INO) {
+        pthread_mutex_lock(&lock);
+        lookup_cnt++;
+        pthread_mutex_unlock(&lock);
+    }
+
     return;
 
 err_out:
@@ -156,9 +177,11 @@ err_out:
 static void tfs_forget (fuse_req_t req, fuse_ino_t ino,
                         uint64_t nlookup) {
     (void) req;
-    if(ino == FILE_INO)
+    if(ino == FILE_INO) {
+        pthread_mutex_lock(&lock);
         lookup_cnt -= nlookup;
-    else
+        pthread_mutex_unlock(&lock);
+    } else
         assert(ino == FUSE_ROOT_ID);
     fuse_reply_none(req);
 }
@@ -229,11 +252,14 @@ static void tfs_open(fuse_req_t req, fuse_ino_t ino,
 
     if (ino == FUSE_ROOT_ID)
         fuse_reply_err(req, EISDIR);
-    else if ((fi->flags & ACCESS_MASK) != O_RDONLY)
+    else if ((fi->flags & O_ACCMODE) != O_RDONLY)
         fuse_reply_err(req, EACCES);
-    else if (ino == FILE_INO)
+    else if (ino == FILE_INO) {
         fuse_reply_open(req, fi);
-    else {
+        pthread_mutex_lock(&lock);
+        open_cnt++;
+        pthread_mutex_unlock(&lock);
+    } else {
         // This should not happen
         fprintf(stderr, "Got open for non-existing inode!\n");
         fuse_reply_err(req, ENOENT);
@@ -274,8 +300,18 @@ static void tfs_retrieve_reply(fuse_req_t req, void *cookie, fuse_ino_t ino,
     fuse_reply_none(req);
 }
 
+static void tfs_destroy(void *userdata)
+{
+	(void)userdata;
 
-static struct fuse_lowlevel_ops tfs_oper = {
+	is_umount = true;
+
+	pthread_join(updater, NULL);
+}
+
+
+static const struct fuse_lowlevel_ops tfs_oper = {
+    .init       = tfs_init,
     .lookup	= tfs_lookup,
     .getattr	= tfs_getattr,
     .readdir	= tfs_readdir,
@@ -283,6 +319,7 @@ static struct fuse_lowlevel_ops tfs_oper = {
     .read	= tfs_read,
     .forget     = tfs_forget,
     .retrieve_reply = tfs_retrieve_reply,
+    .destroy    = tfs_destroy,
 };
 
 static void update_fs(void) {
@@ -302,9 +339,10 @@ static void* update_fs_loop(void *data) {
     struct fuse_bufvec bufv;
     int ret;
 
-    while(1) {
+    while(!is_umount) {
         update_fs();
-        if (!options.no_notify && lookup_cnt) {
+        pthread_mutex_lock(&lock);
+        if (!options.no_notify && open_cnt && lookup_cnt) {
             /* Only send notification if the kernel
                is aware of the inode */
             bufv.count = 1;
@@ -314,15 +352,17 @@ static void* update_fs_loop(void *data) {
             bufv.buf[0].mem = file_contents;
             bufv.buf[0].flags = 0;
 
-            /* This shouldn't fail, but apparenly it sometimes
-               does - see https://github.com/libfuse/libfuse/issues/105 */
+            /*
+             * Some errors (ENOENT, EBADF, ENODEV) have to be accepted as they
+             * might come up during umount, when kernel side already releases
+             * all inodes, but does not send FUSE_DESTROY yet.
+             */
+
             ret = fuse_lowlevel_notify_store(se, FILE_INO, 0, &bufv, 0);
-            if (-ret == ENODEV) {
-                // File system was unmounted
-                break;
-            }
-            else if (ret != 0) {
-                fprintf(stderr, "ERROR: fuse_lowlevel_notify_store() failed with %s (%d)\n",
+            if ((ret != 0 && !is_umount) &&
+                ret != -ENOENT && ret != -EBADF && ret != -ENODEV) {
+                fprintf(stderr,
+                        "ERROR: fuse_lowlevel_notify_store() failed with %s (%d)\n",
                         strerror(-ret), -ret);
                 abort();
             }
@@ -331,13 +371,12 @@ static void* update_fs_loop(void *data) {
                kernel to send us back the stored data */
             ret = fuse_lowlevel_notify_retrieve(se, FILE_INO, MAX_STR_LEN,
                                                 0, (void*) strdup(file_contents));
-            if (-ret == ENODEV) { // File system was unmounted
-                break;
-            }
-            assert(ret == 0);
+            assert((ret == 0 || is_umount) || ret == -ENOENT || ret == -EBADF ||
+                   ret != -ENODEV);
             if(retrieve_status == 0)
                 retrieve_status = 1;
         }
+        pthread_mutex_unlock(&lock);
         sleep(options.update_interval);
     }
     return NULL;
@@ -356,7 +395,7 @@ int main(int argc, char *argv[]) {
     struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
     struct fuse_session *se;
     struct fuse_cmdline_opts opts;
-    pthread_t updater;
+    struct fuse_loop_config *config;
     int ret = -1;
 
     if (fuse_opt_parse(&args, &options, option_spec, NULL) == -1)
@@ -404,8 +443,14 @@ int main(int argc, char *argv[]) {
     /* Block until ctrl+c or fusermount -u */
     if (opts.singlethread)
         ret = fuse_session_loop(se);
-    else
-        ret = fuse_session_loop_mt(se, opts.clone_fd);
+    else {
+	config = fuse_loop_cfg_create();
+	fuse_loop_cfg_set_clone_fd(config, opts.clone_fd);
+	fuse_loop_cfg_set_max_threads(config, opts.max_threads);
+	ret = fuse_session_loop_mt(se, config);
+	fuse_loop_cfg_destroy(config);
+	config = NULL;
+    }
 
     assert(retrieve_status != 1);
     fuse_session_unmount(se);
