@@ -30,12 +30,6 @@
 #define FUSE_MAKE_VERSION(maj, min)  ((maj) * 100 + (min))
 #define FUSE_VERSION FUSE_MAKE_VERSION(FUSE_MAJOR_VERSION, FUSE_MINOR_VERSION)
 
-#ifdef HAVE_STATIC_ASSERT
-#define fuse_static_assert(condition, message) static_assert(condition, message)
-#else
-#define fuse_static_assert(condition, message)
-#endif
-
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -129,8 +123,6 @@ struct fuse_file_info {
 
 	uint64_t reserved[2];
 };
-fuse_static_assert(sizeof(struct fuse_file_info) == 64,
-		   "fuse_file_info size mismatch");
 
 /**
  * Configuration parameters passed to fuse_session_loop_mt() and
@@ -708,8 +700,6 @@ struct fuse_conn_info {
 	 */
 	uint32_t reserved[16];
 };
-fuse_static_assert(sizeof(struct fuse_conn_info) == 128,
-		   "Size of struct fuse_conn_info must be 128 bytes");
 
 struct fuse_session;
 struct fuse_pollhandle;
@@ -1096,28 +1086,40 @@ void fuse_loop_cfg_convert(struct fuse_loop_config *config,
 			   struct fuse_loop_config_v1 *v1_conf);
 #endif
 
+/**
+ * Set a feature flag in the want_ext field of fuse_conn_info.
+ *
+ * @param conn connection information
+ * @param flag feature flag to be set
+ * @return true if the flag was set, false if the flag is not supported
+ */
+bool fuse_set_feature_flag(struct fuse_conn_info *conn, uint64_t flag);
 
-static inline bool fuse_set_feature_flag(struct fuse_conn_info *conn,
-					 uint64_t flag)
-{
-	if (conn->capable_ext & flag) {
-		conn->want_ext |= flag;
-		return true;
-	}
-	return false;
-}
+/**
+ * Unset a feature flag in the want_ext field of fuse_conn_info.
+ *
+ * @param conn connection information
+ * @param flag feature flag to be unset
+ */
+void fuse_unset_feature_flag(struct fuse_conn_info *conn, uint64_t flag);
+
+/**
+ * Get the value of a feature flag in the want_ext field of fuse_conn_info.
+ *
+ * @param conn connection information
+ * @param flag feature flag to be checked
+ * @return true if the flag is set, false otherwise
+ */
+bool fuse_get_feature_flag(struct fuse_conn_info *conn, uint64_t flag);
+
+/*
+ * DO NOT USE: Not part of public API, for internal test use only.
+ * The function signature or any use of it is not guaranteeed to
+ * remain stable. And neither are results of what this function does.
+ */
+int fuse_convert_to_conn_want_ext(struct fuse_conn_info *conn);
 
-static inline void fuse_unset_feature_flag(struct fuse_conn_info *conn,
-					 uint64_t flag)
-{
-	conn->want_ext &= ~flag;
-}
 
-static inline bool fuse_get_feature_flag(struct fuse_conn_info *conn,
-					     uint64_t flag)
-{
-	return conn->capable_ext & flag ? true : false;
-}
 
 /* ----------------------------------------------------------- *
  * Compatibility stuff					       *
