@@ -8,18 +8,16 @@
   See the file COPYING.LIB.
 */
 
-#include "config.h"
+#include "fuse_config.h"
 #include "fuse_i.h"
 #include "fuse_misc.h"
 #include "fuse_opt.h"
+#include "util.h"
 
 #include <sys/param.h>
-#include <sys/mount.h>
+#include "fuse_mount_compat.h"
 
-#include <sys/stat.h>
 #include <sys/wait.h>
-#include <sys/sysctl.h>
-#include <sys/user.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -27,8 +25,6 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
-#include <paths.h>
-#include <limits.h>
 
 #define FUSERMOUNT_PROG		"mount_fusefs"
 #define FUSE_DEV_TRUNK		"/dev/fuse"
@@ -85,7 +81,9 @@ static const struct fuse_opt fuse_mount_opts[] = {
 	FUSE_DUAL_OPT_KEY("private",		KEY_KERN),
 	FUSE_DUAL_OPT_KEY("neglect_shares",	KEY_KERN),
 	FUSE_DUAL_OPT_KEY("push_symlinks_in",	KEY_KERN),
-	FUSE_OPT_KEY("nosync_unmount",		KEY_KERN),
+#if __FreeBSD_version >= 1200519
+	FUSE_DUAL_OPT_KEY("intr",		KEY_KERN),
+#endif
 	/* stock FBSD mountopt parsing routine lets anything be negated... */
 	/*
 	 * Linux specific mount options, but let just the mount util
@@ -108,6 +106,7 @@ unsigned get_max_read(struct mount_opts *o)
 static int fuse_mount_opt_proc(void *data, const char *arg, int key,
 			       struct fuse_args *outargs)
 {
+	(void) outargs;
 	struct mount_opts *mo = data;
 
 	switch (key) {
@@ -125,49 +124,31 @@ static int fuse_mount_opt_proc(void *data, const char *arg, int key,
 
 void fuse_kern_unmount(const char *mountpoint, int fd)
 {
-	char *ep, dev[128];
-	struct stat sbuf;
-	unmount(mountpoint, MNT_FORCE);
-	close(fd);
+	if (close(fd) < 0)
+		fuse_log(FUSE_LOG_ERR, "closing FD %d failed: %s", fd, strerror(errno));
+	if (unmount(mountpoint, MNT_FORCE) < 0)
+		fuse_log(FUSE_LOG_ERR, "unmounting %s failed: %s",
+			mountpoint, strerror(errno));
 }
 
-/* Check if kernel is doing init in background */
-static int init_backgrounded(void)
-{
-	unsigned ibg, len;
-
-	len = sizeof(ibg);
-
-	if (sysctlbyname("vfs.fuse.init_backgrounded", &ibg, &len, NULL, 0))
-		return 0;
-
-	return ibg;
-}
-
-
 static int fuse_mount_core(const char *mountpoint, const char *opts)
 {
 	const char *mountprog = FUSERMOUNT_PROG;
-	int fd;
+	long fd;
 	char *fdnam, *dev;
 	pid_t pid, cpid;
 	int status;
+	int err;
 
 	fdnam = getenv("FUSE_DEV_FD");
 
 	if (fdnam) {
-		char *ep;
-
-		fd = strtol(fdnam, &ep, 10);
-
-		if (*ep != '\0') {
-			fprintf(stderr, "invalid value given in FUSE_DEV_FD\n");
+		err = libfuse_strtol(fdnam, &fd);
+		if (err || fd < 0) {
+			fuse_log(FUSE_LOG_ERR, "invalid value given in FUSE_DEV_FD\n");
 			return -1;
 		}
 
-		if (fd < 0)
-			return -1;
-
 		goto mount;
 	}
 
@@ -195,34 +176,27 @@ mount:
 	}
 
 	if (pid == 0) {
-		if (! init_backgrounded()) {
-			/*
-			 * If init is not backgrounded, we have to
-			 * call the mount util backgrounded, to avoid
-			 * deadlock.
-			 */
+		pid = fork();
 
-			pid = fork();
-
-			if (pid == -1) {
-				perror("fuse: fork() failed");
-				close(fd);
-				exit(1);
-			}
+		if (pid == -1) {
+			perror("fuse: fork() failed");
+			close(fd);
+			_exit(EXIT_FAILURE);
 		}
 
 		if (pid == 0) {
 			const char *argv[32];
 			int a = 0;
 			int ret = -1; 
-			
+
 			if (! fdnam)
 			{
-				ret = asprintf(&fdnam, "%d", fd); 
+				ret = asprintf(&fdnam, "%ld", fd);
 				if(ret == -1)
 				{
 					perror("fuse: failed to assemble mount arguments");
-					exit(1);
+					close(fd);
+					_exit(EXIT_FAILURE);
 				}
 			}
 
@@ -232,23 +206,21 @@ mount:
 				argv[a++] = opts;
 			}
 			argv[a++] = fdnam;
-			
-			if(ret != -1)
-				free(fdnam);
-			
 			argv[a++] = mountpoint;
 			argv[a++] = NULL;
 			execvp(mountprog, (char **) argv);
 			perror("fuse: failed to exec mount program");
-			exit(1);
+			free(fdnam);
+			_exit(EXIT_FAILURE);
 		}
 
-		exit(0);
+		_exit(EXIT_SUCCESS);
 	}
 
 	if (waitpid(cpid, &status, 0) == -1 || WEXITSTATUS(status) != 0) {
 		perror("fuse: failed to mount file system");
-		close(fd);
+		if (close(fd) < 0)
+			perror("fuse: closing FD");
 		return -1;
 	}
 
