@@ -2,12 +2,15 @@
 #define FUSE_UTIL_H_
 
 #include <stdint.h>
+#include <stdbool.h>
 
 #define ROUND_UP(val, round_to) (((val) + (round_to - 1)) & ~(round_to - 1))
 
 #define likely(x) __builtin_expect(!!(x), 1)
 #define unlikely(x) __builtin_expect(!!(x), 0)
 
+struct fuse_conn_info;
+
 int libfuse_strtol(const char *str, long *res);
 
 /**
@@ -30,4 +33,10 @@ static inline uint64_t fuse_higher_32_bits(uint64_t nr)
 #define FUSE_VAR_UNUSED(var) (__attribute__((unused)) var)
 #endif
 
+#define container_of(ptr, type, member)                      \
+	({                                                   \
+		unsigned long __mptr = (unsigned long)(ptr); \
+		((type *)(__mptr - offsetof(type, member))); \
+	})
+
 #endif
