@@ -2,14 +2,17 @@
   FUSE: Filesystem in Userspace
   Copyright (C) 2016 Nikolaus Rath <Nikolaus@rath.org>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
 
 #define FUSE_USE_VERSION 30
 
-#include <config.h>
+/* Not really needed - just to test build with FUSE_USE_VERSION == 30 */
+#include <fuse.h>
+
+#include <fuse_config.h>
 #include <fuse_lowlevel.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -154,7 +157,9 @@ int main(int argc, char *argv[]) {
     pthread_t fs_thread;
 
     assert(fuse_parse_cmdline(&args, &fuse_opts) == 0);
+#ifndef __FreeBSD__    
     assert(fuse_opt_add_arg(&args, "-oauto_unmount") == 0);
+#endif
     se = fuse_session_new(&args, &tfs_oper,
                           sizeof(tfs_oper), NULL);
     assert (se != NULL);
