@@ -1,3 +1,6 @@
+#define _GNU_SOURCE
+#include "fuse_config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -8,34 +11,51 @@
 #include <utime.h>
 #include <errno.h>
 #include <assert.h>
+#include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/un.h>
+
+#ifndef ALLPERMS
+# define ALLPERMS (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)/* 07777 */
+#endif
 
 
+static const char *basepath;
+static const char *basepath_r;
 static char testfile[1024];
 static char testfile2[1024];
 static char testdir[1024];
 static char testdir2[1024];
-static char subfile[1024];
+static char testsock[1024];
+static char subfile[1280];
 
 static char testfile_r[1024];
 static char testfile2_r[1024];
 static char testdir_r[1024];
 static char testdir2_r[1024];
-static char subfile_r[1024];
+static char subfile_r[1280];
 
 static char testname[256];
 static char testdata[] = "abcdefghijklmnopqrstuvwxyz";
 static char testdata2[] = "1234567890-=qwertyuiop[]\asdfghjkl;'zxcvbnm,./";
 static const char *testdir_files[] = { "f1", "f2", NULL};
+static long seekdir_offsets[4];
 static char zerodata[4096];
 static int testdatalen = sizeof(testdata) - 1;
 static int testdata2len = sizeof(testdata2) - 1;
-static unsigned int testnum = 1;
+static unsigned int testnum = 0;
 static unsigned int select_test = 0;
 static unsigned int skip_test = 0;
+static unsigned int unlinked_test = 0;
 
 #define MAX_ENTRIES 1024
+#define MAX_TESTS 100
+
+static struct test {
+	int fd;
+	struct stat stat;
+} tests[MAX_TESTS];
 
 static void test_perror(const char *func, const char *msg)
 {
@@ -59,26 +79,42 @@ static void test_error(const char *func, const char *msg, ...)
 	fprintf(stderr, "\n");
 }
 
+static int is_dot_or_dotdot(const char *name) {
+    return name[0] == '.' &&
+           (name[1] == '\0' || (name[1] == '.' && name[2] == '\0'));
+}
+
 static void success(void)
 {
 	fprintf(stderr, "%s OK\n", testname);
 }
 
+#define this_test (&tests[testnum-1])
+#define next_test (&tests[testnum])
+
 static void __start_test(const char *fmt, ...)
 {
 	unsigned int n;
 	va_list ap;
-	n = sprintf(testname, "%3i [", testnum++);
+	n = sprintf(testname, "%3i [", testnum);
 	va_start(ap, fmt);
 	n += vsprintf(testname + n, fmt, ap);
 	va_end(ap);
 	sprintf(testname + n, "]");
+	// Use dedicated testfile per test
+	sprintf(testfile, "%s/testfile.%d", basepath, testnum);
+	sprintf(testfile_r, "%s/testfile.%d", basepath_r, testnum);
+	if (testnum > MAX_TESTS) {
+		fprintf(stderr, "%s - too many tests\n", testname);
+		exit(1);
+	}
+	this_test->fd = -1;
 }
 
 #define start_test(msg, args...) { \
+	testnum++; \
 	if ((select_test && testnum != select_test) || \
 	    (testnum == skip_test)) { \
-		testnum++; \
 		return 0; \
 	} \
 	__start_test(msg, ##args);		\
@@ -87,6 +123,18 @@ static void __start_test(const char *fmt, ...)
 #define PERROR(msg) test_perror(__FUNCTION__, msg)
 #define ERROR(msg, args...) test_error(__FUNCTION__, msg, ##args)
 
+#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
+
+static int st_check_size(struct stat *st, int len)
+{
+	if (st->st_size != len) {
+		ERROR("length %u instead of %u", (int) st->st_size,
+		      (int) len);
+		return -1;
+	}
+	return 0;
+}
+
 static int check_size(const char *path, int len)
 {
 	struct stat stbuf;
@@ -95,25 +143,19 @@ static int check_size(const char *path, int len)
 		PERROR("stat");
 		return -1;
 	}
-	if (stbuf.st_size != len) {
-		ERROR("length %u instead of %u", (int) stbuf.st_size,
-		      (int) len);
-		return -1;
-	}
-	return 0;
+	return st_check_size(&stbuf, len);
 }
 
-static int fcheck_size(int fd, int len)
+static int check_testfile_size(const char *path, int len)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if (stbuf.st_size != len) {
-		ERROR("length %u instead of %u", (int) stbuf.st_size,
-		      (int) len);
+	this_test->stat.st_size = len;
+	return check_size(path, len);
+}
+
+static int st_check_type(struct stat *st, mode_t type)
+{
+	if ((st->st_mode & S_IFMT) != type) {
+		ERROR("type 0%o instead of 0%o", st->st_mode & S_IFMT, type);
 		return -1;
 	}
 	return 0;
@@ -127,23 +169,14 @@ static int check_type(const char *path, mode_t type)
 		PERROR("lstat");
 		return -1;
 	}
-	if ((stbuf.st_mode & S_IFMT) != type) {
-		ERROR("type 0%o instead of 0%o", stbuf.st_mode & S_IFMT, type);
-		return -1;
-	}
-	return 0;
+	return st_check_type(&stbuf, type);
 }
 
-static int fcheck_type(int fd, mode_t type)
+static int st_check_mode(struct stat *st, mode_t mode)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if ((stbuf.st_mode & S_IFMT) != type) {
-		ERROR("type 0%o instead of 0%o", stbuf.st_mode & S_IFMT, type);
+	if ((st->st_mode & ALLPERMS) != mode) {
+		ERROR("mode 0%o instead of 0%o", st->st_mode & ALLPERMS,
+		      mode);
 		return -1;
 	}
 	return 0;
@@ -157,26 +190,14 @@ static int check_mode(const char *path, mode_t mode)
 		PERROR("lstat");
 		return -1;
 	}
-	if ((stbuf.st_mode & 07777) != mode) {
-		ERROR("mode 0%o instead of 0%o", stbuf.st_mode & 07777, mode);
-		return -1;
-	}
-	return 0;
+	return st_check_mode(&stbuf, mode);
 }
 
-static int fcheck_mode(int fd, mode_t mode)
+static int check_testfile_mode(const char *path, mode_t mode)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if ((stbuf.st_mode & 07777) != mode) {
-		ERROR("mode 0%o instead of 0%o", stbuf.st_mode & 07777, mode);
-		return -1;
-	}
-	return 0;
+	this_test->stat.st_mode &= ~ALLPERMS;
+	this_test->stat.st_mode |= mode;
+	return check_mode(path, mode);
 }
 
 static int check_times(const char *path, time_t atime, time_t mtime)
@@ -227,6 +248,16 @@ static int fcheck_times(int fd, time_t atime, time_t mtime)
 }
 #endif
 
+static int st_check_nlink(struct stat *st, nlink_t nlink)
+{
+	if (st->st_nlink != nlink) {
+		ERROR("nlink %li instead of %li", (long) st->st_nlink,
+		      (long) nlink);
+		return -1;
+	}
+	return 0;
+}
+
 static int check_nlink(const char *path, nlink_t nlink)
 {
 	struct stat stbuf;
@@ -235,28 +266,32 @@ static int check_nlink(const char *path, nlink_t nlink)
 		PERROR("lstat");
 		return -1;
 	}
-	if (stbuf.st_nlink != nlink) {
-		ERROR("nlink %li instead of %li", (long) stbuf.st_nlink,
-		      (long) nlink);
-		return -1;
-	}
-	return 0;
+	return st_check_nlink(&stbuf, nlink);
 }
 
-static int fcheck_nlink(int fd, nlink_t nlink)
+static int fcheck_stat(int fd, int flags, struct stat *st)
 {
 	struct stat stbuf;
 	int res = fstat(fd, &stbuf);
 	if (res == -1) {
+		if (flags & O_PATH) {
+			// With O_PATH fd, the server does not have to keep
+			// the inode alive so FUSE inode may be stale or bad
+			if (errno == ESTALE || errno == EIO ||
+			    errno == ENOENT || errno == EBADF)
+				return 0;
+		}
 		PERROR("fstat");
 		return -1;
 	}
-	if (stbuf.st_nlink != nlink) {
-		ERROR("nlink %li instead of %li", (long) stbuf.st_nlink,
-		      (long) nlink);
-		return -1;
-	}
-	return 0;
+
+	int err = 0;
+	err += st_check_type(&stbuf, st->st_mode & S_IFMT);
+	err += st_check_mode(&stbuf, st->st_mode & ALLPERMS);
+	err += st_check_size(&stbuf, st->st_size);
+	err += st_check_nlink(&stbuf, st->st_nlink);
+
+	return err;
 }
 
 static int check_nonexist(const char *path)
@@ -369,10 +404,6 @@ static int check_dir_contents(const char *path, const char **contents)
 		found[i] = 0;
 		cont[i] = contents[i];
 	}
-	found[i] = 0;
-	cont[i++] = ".";
-	found[i] = 0;
-	cont[i++] = "..";
 	cont[i] = NULL;
 
 	dp = opendir(path);
@@ -393,6 +424,8 @@ static int check_dir_contents(const char *path, const char **contents)
 			}
 			break;
 		}
+		if (is_dot_or_dotdot(de->d_name))
+			continue;
 		for (i = 0; cont[i] != NULL; i++) {
 			assert(i < MAX_ENTRIES);
 			if (strcmp(cont[i], de->d_name) == 0) {
@@ -478,6 +511,90 @@ static int create_file(const char *path, const char *data, int len)
 	return 0;
 }
 
+static int create_path_fd(const char *path, const char *data, int len)
+{
+	int path_fd;
+	int res;
+
+	res = create_file(path, data, len);
+	if (res == -1)
+		return -1;
+
+	path_fd = open(path, O_PATH);
+	if (path_fd == -1)
+		PERROR("open(O_PATH)");
+
+	return path_fd;
+}
+
+// Can be called once per test
+static int create_testfile(const char *path, const char *data, int len)
+{
+	struct test *t = this_test;
+	struct stat *st = &t->stat;
+	int res, fd;
+
+	if (t->fd > 0) {
+		ERROR("testfile already created");
+		return -1;
+	}
+
+	fd = create_path_fd(path, data, len);
+	if (fd == -1)
+		return -1;
+
+	t->fd = fd;
+
+	res = fstat(fd, st);
+	if (res == -1) {
+		PERROR("fstat");
+		return -1;
+	}
+
+	return 0;
+}
+
+static int check_unlinked_testfile(int fd)
+{
+	struct stat *st = &this_test->stat;
+
+	st->st_nlink = 0;
+	return fcheck_stat(fd, O_PATH, st);
+}
+
+// Check recorded testfiles after all tests completed
+static int check_unlinked_testfiles(void)
+{
+	int fd;
+	int res, err = 0;
+	int num = testnum;
+
+	if (!unlinked_test)
+		return 0;
+
+	testnum = 0;
+	while (testnum < num) {
+		fd = next_test->fd;
+		start_test("check_unlinked_testfile");
+		if (fd == -1)
+			continue;
+
+		err += check_unlinked_testfile(fd);
+		res = close(fd);
+		if (res == -1) {
+			PERROR("close(test_fd)");
+			err--;
+		}
+	}
+
+	if (err) {
+		fprintf(stderr, "%i unlinked testfile checks failed\n", -err);
+		return 1;
+	}
+
+	return err;
+}
+
 static int cleanup_dir(const char *path, const char **dir_files, int quiet)
 {
 	int i;
@@ -485,7 +602,7 @@ static int cleanup_dir(const char *path, const char **dir_files, int quiet)
 
 	for (i = 0; dir_files[i]; i++) {
 		int res;
-		char fpath[1024];
+		char fpath[1280];
 		sprintf(fpath, "%s/%s", path, dir_files[i]);
 		res = unlink(fpath);
 		if (res == -1 && !quiet) {
@@ -518,7 +635,7 @@ static int create_dir(const char *path, const char **dir_files)
 		return -1;
 
 	for (i = 0; dir_files[i]; i++) {
-		char fpath[1024];
+		char fpath[1280];
 		sprintf(fpath, "%s/%s", path, dir_files[i]);
 		res = create_file(fpath, "", 0);
 		if (res == -1) {
@@ -542,7 +659,7 @@ static int test_truncate(int len)
 	int res;
 
 	start_test("truncate(%u)", (int) len);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -551,7 +668,7 @@ static int test_truncate(int len)
 		PERROR("truncate");
 		return -1;
 	}
-	res = check_size(testfile, len);
+	res = check_testfile_size(testfile, len);
 	if (res == -1)
 		return -1;
 
@@ -591,7 +708,7 @@ static int test_ftruncate(int len, int mode)
 	int fd;
 
 	start_test("ftruncate(%u) mode: 0%03o", len, mode);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -607,7 +724,7 @@ static int test_ftruncate(int len, int mode)
 		close(fd);
 		return -1;
 	}
-	res = check_mode(testfile, mode);
+	res = check_testfile_mode(testfile, mode);
 	if (res == -1) {
 		close(fd);
 		return -1;
@@ -619,7 +736,7 @@ static int test_ftruncate(int len, int mode)
 		return -1;
 	}
 	close(fd);
-	res = check_size(testfile, len);
+	res = check_testfile_size(testfile, len);
 	if (res == -1)
 		return -1;
 
@@ -651,6 +768,159 @@ static int test_ftruncate(int len, int mode)
 	return 0;
 }
 
+static int test_seekdir(void)
+{
+	int i;
+	int res;
+	DIR *dp;
+	struct dirent *de = NULL;
+
+	start_test("seekdir");
+	res = create_dir(testdir, testdir_files);
+	if (res == -1)
+		return res;
+
+	dp = opendir(testdir);
+	if (dp == NULL) {
+		PERROR("opendir");
+		return -1;
+	}
+
+	/* Remember dir offsets */
+	for (i = 0; i < ARRAY_SIZE(seekdir_offsets); i++) {
+		seekdir_offsets[i] = telldir(dp);
+		errno = 0;
+		de = readdir(dp);
+		if (de == NULL) {
+			if (errno) {
+				PERROR("readdir");
+				goto fail;
+			}
+			break;
+		}
+	}
+
+	/* Walk until the end of directory */
+	while (de)
+		de = readdir(dp);
+
+	/* Start from the last valid dir offset and seek backwards */
+	for (i--; i >= 0; i--) {
+		seekdir(dp, seekdir_offsets[i]);
+		de = readdir(dp);
+		if (de == NULL) {
+			ERROR("Unexpected end of directory after seekdir()");
+			goto fail;
+		}
+	}
+
+	closedir(dp);
+	res = cleanup_dir(testdir, testdir_files, 0);
+	if (!res)
+		success();
+	return res;
+fail:
+	closedir(dp);
+	cleanup_dir(testdir, testdir_files, 1);
+	return -1;
+}
+
+#ifdef HAVE_COPY_FILE_RANGE
+static int test_copy_file_range(void)
+{
+	const char *data = testdata;
+	int datalen = testdatalen;
+	int err = 0;
+	int res;
+	int fd_in, fd_out;
+	off_t pos_in = 0, pos_out = 0;
+
+	start_test("copy_file_range");
+	unlink(testfile);
+	fd_in = open(testfile, O_CREAT | O_RDWR, 0644);
+	if (fd_in == -1) {
+		PERROR("creat");
+		return -1;
+	}
+	res = write(fd_in, data, datalen);
+	if (res == -1) {
+		PERROR("write");
+		close(fd_in);
+		return -1;
+	}
+	if (res != datalen) {
+		ERROR("write is short: %u instead of %u", res, datalen);
+		close(fd_in);
+		return -1;
+	}
+
+	unlink(testfile2);
+	fd_out = creat(testfile2, 0644);
+	if (fd_out == -1) {
+		PERROR("creat");
+		close(fd_in);
+		return -1;
+	}
+	res = copy_file_range(fd_in, &pos_in, fd_out, &pos_out, datalen, 0);
+	if (res == -1) {
+		PERROR("copy_file_range");
+		close(fd_in);
+		close(fd_out);
+		return -1;
+	}
+	if (res != datalen) {
+		ERROR("copy is short: %u instead of %u", res, datalen);
+		close(fd_in);
+		close(fd_out);
+		return -1;
+	}
+
+	res = close(fd_in);
+	if (res == -1) {
+		PERROR("close");
+		close(fd_out);
+		return -1;
+	}
+	res = close(fd_out);
+	if (res == -1) {
+		PERROR("close");
+		return -1;
+	}
+
+	err = check_data(testfile2, data, 0, datalen);
+
+	res = unlink(testfile);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile);
+	if (res == -1)
+		return -1;
+	if (err)
+		return -1;
+
+	res = unlink(testfile2);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile2);
+	if (res == -1)
+		return -1;
+	if (err)
+		return -1;
+
+	success();
+	return 0;
+}
+#else
+static int test_copy_file_range(void)
+{
+	return 0;
+}
+#endif
+
 static int test_utime(void)
 {
 	struct utimbuf utm;
@@ -659,7 +929,7 @@ static int test_utime(void)
 	int res;
 
 	start_test("utime");
-	res = create_file(testfile, NULL, 0);
+	res = create_testfile(testfile, NULL, 0);
 	if (res == -1)
 		return -1;
 
@@ -762,8 +1032,10 @@ static int test_create_unlink(void)
 		return -1;
 	}
 	res = check_nonexist(testfile);
-	if (res == -1)
+	if (res == -1) {
+		close(fd);
 		return -1;
+	}
 	res = write(fd, data, datalen);
 	if (res == -1) {
 		PERROR("write");
@@ -775,10 +1047,11 @@ static int test_create_unlink(void)
 		close(fd);
 		return -1;
 	}
-	err += fcheck_type(fd, S_IFREG);
-	err += fcheck_mode(fd, 0644);
-	err += fcheck_nlink(fd, 0);
-	err += fcheck_size(fd, datalen);
+	struct stat st = {
+		.st_mode = S_IFREG | 0644,
+		.st_size = datalen,
+	};
+	err = fcheck_stat(fd, O_RDWR, &st);
 	err += fcheck_data(fd, data, 0, datalen);
 	res = close(fd);
 	if (res == -1) {
@@ -880,7 +1153,7 @@ static int do_test_open(int exist, int flags, const char *flags_str, int mode)
 		err += check_mode(testfile, mode);
 	err += check_nlink(testfile, 1);
 	err += check_size(testfile, currlen);
-	if (exist && !(flags & O_TRUNC) && (mode & 0400))
+	if (exist && !(flags & O_TRUNC) && (mode & S_IRUSR))
 		err += check_data(testfile, testdata2, 0, testdata2len);
 
 	res = write(fd, data, datalen);
@@ -897,7 +1170,7 @@ static int do_test_open(int exist, int flags, const char *flags_str, int mode)
 
 			err += check_size(testfile, currlen);
 
-			if (mode & 0400) {
+			if (mode & S_IRUSR) {
 				err += check_data(testfile, data, 0, datalen);
 				if (exist && !(flags & O_TRUNC) &&
 				    testdata2len > datalen)
@@ -996,7 +1269,7 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 	start_test("open_acc(%s) mode: 0%03o message: '%s'", flags_str, mode,
 		   strerror(err));
 	unlink(testfile);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1006,7 +1279,7 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 		return -1;
 	}
 
-	res = check_mode(testfile, mode);
+	res = check_testfile_mode(testfile, mode);
 	if (res == -1)
 		return -1;
 
@@ -1024,6 +1297,19 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 		}
 		close(fd);
 	}
+
+	res = unlink(testfile);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile);
+	if (res == -1)
+		return -1;
+	res = check_nonexist(testfile_r);
+	if (res == -1)
+		return -1;
+
 	success();
 	return 0;
 }
@@ -1038,7 +1324,7 @@ static int test_symlink(void)
 	int res;
 
 	start_test("symlink");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1079,6 +1365,15 @@ static int test_symlink(void)
 	if (err)
 		return -1;
 
+	res = unlink(testfile);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile);
+	if (res == -1)
+		return -1;
+
 	success();
 	return 0;
 }
@@ -1091,7 +1386,7 @@ static int test_link(void)
 	int res;
 
 	start_test("link");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1141,7 +1436,7 @@ static int test_link2(void)
 	int res;
 
 	start_test("link-unlink-link");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1200,7 +1495,7 @@ static int test_rename_file(void)
 	int res;
 
 	start_test("rename file");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1280,6 +1575,221 @@ static int test_rename_dir(void)
 	return 0;
 }
 
+static int test_rename_dir_loop(void)
+{
+#define PATH(p)		(snprintf(path, sizeof path, "%s/%s", testdir, p), path)
+#define PATH2(p)	(snprintf(path2, sizeof path2, "%s/%s", testdir, p), path2)
+
+	char path[1280], path2[1280];
+	int err = 0;
+	int res;
+
+	start_test("rename dir loop");
+
+	res = create_dir(testdir, testdir_files);
+	if (res == -1)
+		return -1;
+
+	res = mkdir(PATH("a"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	res = rename(PATH("a"), PATH2("a"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	errno = 0;
+	res = rename(PATH("a"), PATH2("a/b"));
+	if (res == 0 || errno != EINVAL) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = mkdir(PATH("a/b"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	res = mkdir(PATH("a/b/c"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	errno = 0;
+	res = rename(PATH("a"), PATH2("a/b/c"));
+	if (res == 0 || errno != EINVAL) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	errno = 0;
+	res = rename(PATH("a"), PATH2("a/b/c/a"));
+	if (res == 0 || errno != EINVAL) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	errno = 0;
+	res = rename(PATH("a/b/c"), PATH2("a"));
+	if (res == 0 || errno != ENOTEMPTY) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = open(PATH("a/foo"), O_CREAT, 0644);
+	if (res == -1) {
+		PERROR("open");
+		goto fail;
+	}
+	close(res);
+
+	res = rename(PATH("a/foo"), PATH2("a/bar"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/bar"), PATH2("a/foo"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/foo"), PATH2("a/b/bar"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/b/bar"), PATH2("a/foo"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/foo"), PATH2("a/b/c/bar"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/b/c/bar"), PATH2("a/foo"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = open(PATH("a/bar"), O_CREAT, 0644);
+	if (res == -1) {
+		PERROR("open");
+		goto fail;
+	}
+	close(res);
+
+	res = rename(PATH("a/foo"), PATH2("a/bar"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	unlink(PATH("a/bar"));
+
+	res = rename(PATH("a/b"), PATH2("a/d"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/d"), PATH2("a/b"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = mkdir(PATH("a/d"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	res = rename(PATH("a/b"), PATH2("a/d"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = rename(PATH("a/d"), PATH2("a/b"));
+	if (res == -1) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	res = mkdir(PATH("a/d"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	res = mkdir(PATH("a/d/e"), 0755);
+	if (res == -1) {
+		PERROR("mkdir");
+		goto fail;
+	}
+
+	errno = 0;
+	res = rename(PATH("a/b"), PATH2("a/d"));
+	if (res == 0 || (errno != ENOTEMPTY && errno != EEXIST)) {
+		PERROR("rename");
+		goto fail;
+	}
+
+	rmdir(PATH("a/d/e"));
+	rmdir(PATH("a/d"));
+
+ 	rmdir(PATH("a/b/c"));
+	rmdir(PATH("a/b"));
+	rmdir(PATH("a"));
+
+	err += cleanup_dir(testdir, testdir_files, 0);
+	res = rmdir(testdir);
+	if (res == -1) {
+		PERROR("rmdir");
+		goto fail;
+	}
+	res = check_nonexist(testdir);
+	if (res == -1)
+		return -1;
+	if (err)
+		return -1;
+
+	success();
+	return 0;
+
+fail:
+	unlink(PATH("a/bar"));
+
+	rmdir(PATH("a/d/e"));
+	rmdir(PATH("a/d"));
+ 
+ 	rmdir(PATH("a/b/c"));
+	rmdir(PATH("a/b"));
+	rmdir(PATH("a"));
+
+	cleanup_dir(testdir, testdir_files, 1);
+	rmdir(testdir);
+
+	return -1;
+
+#undef PATH2
+#undef PATH
+}
+
 static int test_mkfifo(void)
 {
 	int res;
@@ -1329,7 +1839,9 @@ static int test_mkdir(void)
 	if (res == -1)
 		return -1;
 	err += check_mode(testdir, 0755);
-	err += check_nlink(testdir, 2);
+	/* Some file systems (like btrfs) don't track link
+	   count for directories */
+	//err += check_nlink(testdir, 2);
 	err += check_dir_contents(testdir, dir_contents);
 	res = rmdir(testdir);
 	if (res == -1) {
@@ -1346,6 +1858,58 @@ static int test_mkdir(void)
 	return 0;
 }
 
+static int test_socket(void)
+{
+	struct sockaddr_un su;
+	int fd;
+	int res;
+	int err = 0;
+    const size_t test_sock_len = strlen(testsock) + 1;
+
+	start_test("socket");
+	if (test_sock_len > sizeof(su.sun_path)) {
+		fprintf(stderr, "Need to shorten mount point by %zu chars\n",
+			strlen(testsock) + 1 - sizeof(su.sun_path));
+		return -1;
+	}
+	unlink(testsock);
+	fd = socket(AF_UNIX, SOCK_STREAM, 0);
+	if (fd < 0) {
+		PERROR("socket");
+		return -1;
+	}
+	su.sun_family = AF_UNIX;
+
+	strncpy(su.sun_path, testsock, test_sock_len);
+	su.sun_path[sizeof(su.sun_path) - 1] = '\0';
+	res = bind(fd, (struct sockaddr*)&su, sizeof(su));
+	if (res == -1) {
+		PERROR("bind");
+		return -1;
+	}
+
+	res = check_type(testsock, S_IFSOCK);
+	if (res == -1) {
+		close(fd);
+		return -1;
+	}
+	err += check_nlink(testsock, 1);
+	close(fd);
+	res = unlink(testsock);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testsock);
+	if (res == -1)
+		return -1;
+	if (err)
+		return -1;
+
+	success();
+	return 0;
+}
+
 #define test_create_ro_dir(flags)	 \
 	do_test_create_ro_dir(flags, #flags)
 
@@ -1389,41 +1953,144 @@ static int do_test_create_ro_dir(int flags, const char *flags_str)
 	return 0;
 }
 
+#ifndef __FreeBSD__
+/* 	this tests open with O_TMPFILE
+	note that this will only work with the fuse low level api 
+	you will get ENOTSUP with the high level api */
+static int test_create_tmpfile(void) 
+{
+	rmdir(testdir);
+	int res = mkdir(testdir, 0777);
+	if (res)
+		return -1;
+	
+	start_test("create tmpfile");
+
+	int fd = open(testdir, O_TMPFILE | O_RDWR, S_IRUSR | S_IWUSR);
+	if(fd == -1) {
+		if (errno == ENOTSUP) {
+			/* don't bother if we're working on an old kernel 
+					or on the high level API */
+			return 0;
+		}
+
+		PERROR("open O_TMPFILE | O_RDWR");
+		return -1;
+	}
+	close(fd);
+
+	fd = open(testdir, O_TMPFILE | O_WRONLY | O_EXCL, S_IRUSR | S_IWUSR);
+	if(fd == -1){
+		PERROR("open with O_TMPFILE | O_WRONLY | O_EXCL");
+		return -1;
+	};
+	close(fd);
+
+	fd = open(testdir, O_TMPFILE | O_RDONLY, S_IRUSR);
+	if (fd != -1) {
+		ERROR("open with O_TMPFILE | O_RDONLY succeeded");
+		return -1;
+	}
+	
+	success();
+	return 0;	
+}
+
+static int test_create_and_link_tmpfile(void) 
+{
+	/* skip this test for now since the github runner will fail in the linkat call below */
+	return 0;
+
+	rmdir(testdir);
+	unlink(testfile);
+
+	int res = mkdir(testdir, 0777);
+	if (res)
+		return -1;
+
+	start_test("create and link tmpfile");
+
+	int fd = open(testdir, O_TMPFILE | O_RDWR | O_EXCL, S_IRUSR | S_IWUSR);
+	if(fd == -1) {
+		if (errno == ENOTSUP) {
+			/* don't bother if we're working on an old kernel
+				or on the high level API */
+			return 0;
+		}
+		PERROR("open with O_TMPFILE | O_RDWR | O_EXCL");
+		return -1;
+	}
+
+	if (!linkat(fd, "", AT_FDCWD, testfile, AT_EMPTY_PATH)) {
+		ERROR("linkat succeeded on a tmpfile opened with O_EXCL");
+		return -1;
+	}
+	close(fd);
+
+	fd = open(testdir, O_TMPFILE | O_RDWR, S_IRUSR | S_IWUSR);
+	if(fd == -1) {
+		PERROR("open O_TMPFILE");
+		return -1;
+	}
+	
+	if (check_nonexist(testfile)) {
+		return -1;
+	}
+
+	if (linkat(fd, "", AT_FDCWD, testfile, AT_EMPTY_PATH)) {
+		PERROR("linkat tempfile");
+		return -1;
+	}
+	close(fd);
+
+	if (check_nlink(testfile, 1)) {
+		return -1;
+	}
+	unlink(testfile);
+
+	success();
+	return 0;
+}
+#endif
+
 int main(int argc, char *argv[])
 {
-	const char *basepath;
-	const char *realpath;
 	int err = 0;
 	int a;
 	int is_root;
 
 	umask(0);
 	if (argc < 2 || argc > 4) {
-		fprintf(stderr, "usage: %s testdir [:realdir] [[-]test#]\n", argv[0]);
+		fprintf(stderr, "usage: %s testdir [:realdir] [[-]test#] [-u]\n", argv[0]);
 		return 1;
 	}
 	basepath = argv[1];
-	realpath = basepath;
+	basepath_r = basepath;
 	for (a = 2; a < argc; a++) {
 		char *endptr;
 		char *arg = argv[a];
 		if (arg[0] == ':') {
-			realpath = arg + 1;
+			basepath_r = arg + 1;
 		} else {
 			if (arg[0] == '-') {
 				arg++;
-				skip_test = strtoul(arg, &endptr, 10);
+				if (arg[0] == 'u') {
+					unlinked_test = 1;
+					endptr = arg + 1;
+				} else {
+					skip_test = strtoul(arg, &endptr, 10);
+				}
 			} else {
 				select_test = strtoul(arg, &endptr, 10);
 			}
 			if (arg[0] == '\0' || *endptr != '\0') {
-				fprintf(stderr, "invalid number: '%s'\n", arg);
+				fprintf(stderr, "invalid option: '%s'\n", argv[a]);
 				return 1;
 			}
 		}
 	}
 	assert(strlen(basepath) < 512);
-	assert(strlen(realpath) < 512);
+	assert(strlen(basepath_r) < 512);
 	if (basepath[0] != '/') {
 		fprintf(stderr, "testdir must be an absolute path\n");
 		return 1;
@@ -1434,25 +2101,29 @@ int main(int argc, char *argv[])
 	sprintf(testdir, "%s/testdir", basepath);
 	sprintf(testdir2, "%s/testdir2", basepath);
 	sprintf(subfile, "%s/subfile", testdir2);
+	sprintf(testsock, "%s/testsock", basepath);
 
-	sprintf(testfile_r, "%s/testfile", realpath);
-	sprintf(testfile2_r, "%s/testfile2", realpath);
-	sprintf(testdir_r, "%s/testdir", realpath);
-	sprintf(testdir2_r, "%s/testdir2", realpath);
+	sprintf(testfile_r, "%s/testfile", basepath_r);
+	sprintf(testfile2_r, "%s/testfile2", basepath_r);
+	sprintf(testdir_r, "%s/testdir", basepath_r);
+	sprintf(testdir2_r, "%s/testdir2", basepath_r);
 	sprintf(subfile_r, "%s/subfile", testdir2_r);
 
 	is_root = (geteuid() == 0);
 
 	err += test_create();
 	err += test_create_unlink();
-	err += test_mknod();
 	err += test_symlink();
 	err += test_link();
 	err += test_link2();
+	err += test_mknod();
 	err += test_mkfifo();
 	err += test_mkdir();
 	err += test_rename_file();
 	err += test_rename_dir();
+	err += test_rename_dir_loop();
+	err += test_seekdir();
+	err += test_socket();
 	err += test_utime();
 	err += test_truncate(0);
 	err += test_truncate(testdatalen / 2);
@@ -1507,9 +2178,14 @@ int main(int argc, char *argv[])
 	err += test_create_ro_dir(O_CREAT | O_EXCL);
 	err += test_create_ro_dir(O_CREAT | O_WRONLY);
 	err += test_create_ro_dir(O_CREAT | O_TRUNC);
+	err += test_copy_file_range();
+#ifndef __FreeBSD__
+	err += test_create_tmpfile();
+	err += test_create_and_link_tmpfile();
+#endif
 
-	unlink(testfile);
 	unlink(testfile2);
+	unlink(testsock);
 	rmdir(testdir);
 	rmdir(testdir2);
 
@@ -1518,5 +2194,5 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
-	return 0;
+	return check_unlinked_testfiles();
 }
