@@ -2,14 +2,17 @@
   FUSE: Filesystem in Userspace
   Copyright (C) 2016 Nikolaus Rath <Nikolaus@rath.org>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
 
 #define FUSE_USE_VERSION 30
 
-#include <config.h>
+/* Not really needed - just to test build with FUSE_USE_VERSION == 30 */
+#include <fuse.h>
+
+#include <fuse_config.h>
 #include <fuse_lowlevel.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -19,7 +22,9 @@
 #include <assert.h>
 #include <stddef.h>
 #include <unistd.h>
+#include <sys/stat.h>
 #include <pthread.h>
+#include <stdatomic.h>
 
 #ifndef __linux__
 #include <limits.h>
@@ -34,27 +39,37 @@
 struct options {
     int writeback;
     int data_size;
+    int delay_ms;
 } options = {
     .writeback = 0,
-    .data_size = 4096,
+    .data_size = 2048,
+    .delay_ms = 0,
 };
 
+#define WRITE_SYSCALLS 64
+
 #define OPTION(t, p)                           \
     { t, offsetof(struct options, p), 1 }
 static const struct fuse_opt option_spec[] = {
     OPTION("writeback_cache", writeback),
     OPTION("--data-size=%d", data_size),
+    OPTION("--delay_ms=%d", delay_ms),
     FUSE_OPT_END
 };
 static int got_write;
+static atomic_int write_cnt;
+
+pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
+pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
+static int write_start, write_done;
 
 static void tfs_init (void *userdata, struct fuse_conn_info *conn)
 {
     (void) userdata;
 
     if(options.writeback) {
-        assert(conn->capable & FUSE_CAP_WRITEBACK_CACHE);
-        conn->want |= FUSE_CAP_WRITEBACK_CACHE;
+        assert(fuse_get_feature_flag(conn, FUSE_CAP_WRITEBACK_CACHE));
+        fuse_set_feature_flag(conn, FUSE_CAP_WRITEBACK_CACHE);
     }
 }
 
@@ -117,6 +132,9 @@ static void tfs_open(fuse_req_t req, fuse_ino_t ino,
         fuse_reply_err(req, EISDIR);
     else {
         assert(ino == FILE_INO);
+        /* Test close(rofd) does not block waiting for pending writes */
+        fi->noflush = !options.writeback && options.delay_ms &&
+                      (fi->flags & O_ACCMODE) == O_RDONLY;
         fuse_reply_open(req, fi);
     }
 }
@@ -131,11 +149,29 @@ static void tfs_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
     if(options.writeback)
         expected *= 2;
 
-    if(size != expected)
-        fprintf(stderr, "ERROR: Expected %zd bytes, got %zd\n!",
-                expected, size);
-    else
-        got_write = 1;
+    write_cnt++;
+
+   if(size != expected && !options.writeback)
+       fprintf(stderr, "ERROR: Expected %zd bytes, got %zd\n!",
+               expected, size);
+   else
+      got_write = 1;
+
+    /* Simulate waiting for pending writes */
+    if (options.delay_ms) {
+        pthread_mutex_lock(&lock);
+        write_start = 1;
+        pthread_cond_signal(&cond);
+        pthread_mutex_unlock(&lock);
+
+        usleep(options.delay_ms * 1000);
+
+        pthread_mutex_lock(&lock);
+        write_done = 1;
+        pthread_cond_signal(&cond);
+        pthread_mutex_unlock(&lock);
+    }
+
     fuse_reply_write(req, size);
 }
 
@@ -147,6 +183,25 @@ static struct fuse_lowlevel_ops tfs_oper = {
     .write	= tfs_write,
 };
 
+static void* close_rofd(void *data) {
+    int rofd = (int)(long) data;
+
+    /* Wait for first write to start */
+    pthread_mutex_lock(&lock);
+    while (!write_start && !write_done)
+        pthread_cond_wait(&cond, &lock);
+    pthread_mutex_unlock(&lock);
+
+    close(rofd);
+    printf("rofd closed. write_start: %d write_done: %d\n", write_start, write_done);
+
+    /* First write should not have been completed */
+    if (write_done)
+        fprintf(stderr, "ERROR: close(rofd) blocked on write!\n");
+
+    return NULL;
+}
+
 static void* run_fs(void *data) {
     struct fuse_session *se = (struct fuse_session*) data;
     assert(fuse_session_loop(se) == 0);
@@ -156,8 +211,11 @@ static void* run_fs(void *data) {
 static void test_fs(char *mountpoint) {
     char fname[PATH_MAX];
     char *buf;
-    size_t dsize = options.data_size;
-    int fd;
+    const size_t iosize = options.data_size;
+    const size_t dsize = options.data_size * WRITE_SYSCALLS;
+    int fd, rofd;
+    pthread_t rofd_thread;
+    off_t off = 0;
 
     buf = malloc(dsize);
     assert(buf != NULL);
@@ -173,9 +231,26 @@ static void test_fs(char *mountpoint) {
         assert(0);
     }
 
-    assert(write(fd, buf, dsize) == dsize);
-    assert(write(fd, buf, dsize) == dsize);
+    if (options.delay_ms) {
+        /* Verify that close(rofd) does not block waiting for pending writes */
+        rofd = open(fname, O_RDONLY);
+        assert(pthread_create(&rofd_thread, NULL, close_rofd, (void *)(long)rofd) == 0);
+        /* Give close_rofd time to start */
+        usleep(options.delay_ms * 1000);
+    }
+
+    for (int cnt = 0; cnt < WRITE_SYSCALLS; cnt++) {
+        assert(pwrite(fd, buf + off, iosize, off) == iosize);
+        off += iosize;
+        assert(off <= dsize);
+    }
+    free(buf);
     close(fd);
+
+    if (options.delay_ms) {
+        printf("rwfd closed. write_start: %d write_done: %d\n", write_start, write_done);
+        assert(pthread_join(rofd_thread, NULL) == 0);
+    }
 }
 
 int main(int argc, char *argv[]) {
@@ -186,9 +261,12 @@ int main(int argc, char *argv[]) {
 
     assert(fuse_opt_parse(&args, &options, option_spec, NULL) == 0);
     assert(fuse_parse_cmdline(&args, &fuse_opts) == 0);
+#ifndef __FreeBSD__    
     assert(fuse_opt_add_arg(&args, "-oauto_unmount") == 0);
+#endif
     se = fuse_session_new(&args, &tfs_oper,
                           sizeof(tfs_oper), NULL);
+    fuse_opt_free_args(&args);
     assert (se != NULL);
     assert(fuse_set_signal_handlers(se) == 0);
     assert(fuse_session_mount(se, fuse_opts.mountpoint) == 0);
@@ -198,12 +276,28 @@ int main(int argc, char *argv[]) {
 
     /* Write test data */
     test_fs(fuse_opts.mountpoint);
+    free(fuse_opts.mountpoint);
 
     /* Stop file system */
-    assert(pthread_cancel(fs_thread) == 0);
-
+    fuse_session_exit(se);
     fuse_session_unmount(se);
+    assert(pthread_join(fs_thread, NULL) == 0);
+
     assert(got_write == 1);
+
+    /*
+     * when writeback cache is enabled, kernel side can merge requests, but
+     * memory pressure, system 'sync' might trigger data flushes before - flush
+     * might happen in between write syscalls - merging subpage writes into
+     * a single page and pages into large fuse requests might or might not work.
+     * Though we can expect that that at least some (but maybe all) write
+     * system calls can be merged.
+     */
+    if (options.writeback)
+        assert(write_cnt < WRITE_SYSCALLS);
+    else
+        assert(write_cnt == WRITE_SYSCALLS);
+
     fuse_remove_signal_handlers(se);
     fuse_session_destroy(se);
 
